/************************************************************************

	FourScoreUSB.c

	Nintendo Four Score USB Joystick Adapter
	Copyright (C) 2013 Simon Inns

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	Email: simon.inns@gmail.com

************************************************************************/

#include "FourScoreUSB.h"

// Array for storing the physical button state, de-bounce counter and current de-bounced button state
struct buttonState
{
	uint8_t physicalState; // On or off
	uint8_t state; // On or off
	uint8_t debounceCount;
};

// Array for storing the physical state of all four joysticks
struct joystickState
{
	struct buttonState button[NUMBER_OF_BUTTONS];
} joyStick[4];

// Global for storing the previously sent reports for each joystick
USB_JoystickReport_Data_t previousJoystickReportData0;
USB_JoystickReport_Data_t previousJoystickReportData1;
USB_JoystickReport_Data_t previousJoystickReportData2;
USB_JoystickReport_Data_t previousJoystickReportData3;

// Main function
int main(void)
{
	// Set up the hardware and enable interrupts
	SetupHardware();
	GlobalInterruptEnable();

	// Main program loop
	while(1)
	{
		// Read the four score buttons and joystick states
		readButtonStates();
		readJoystickStates();
		
		// Perform button and joystick debouncing
		performDebounce();
		
		// Perform the local HID tasks
		HID_Task();
		
		// Perform the LUFA USB tasks
		USB_USBTask();
	}
}

// Set up the hardware
void SetupHardware(void)
{
	// Disable watchdog if enabled by bootloader/fuses
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	// Disable clock division
	clock_prescale_set(clock_div_1);

	// Hardware Initialization
	
	// Joystick data input
	JOY1_DATA_DDR &= ~(JOY1_DATA);			// Set direction to input
	JOY1_DATA_PORT |= JOY1_DATA;			// Turn on weak pull-ups
	JOY2_DATA_DDR &= ~(JOY2_DATA);			// Set direction to input
	JOY2_DATA_PORT |= JOY2_DATA;			// Turn on weak pull-ups
	JOY3_DATA_DDR &= ~(JOY3_DATA);			// Set direction to input
	JOY3_DATA_PORT |= JOY3_DATA;			// Turn on weak pull-ups
	JOY4_DATA_DDR &= ~(JOY4_DATA);			// Set direction to input
	JOY4_DATA_PORT |= JOY4_DATA;			// Turn on weak pull-ups
	
	// Joystick clock output
	JOY_CLOCK_DDR |= JOY_CLOCK;				// Set direction to output
	JOY_CLOCK_PORT &= ~(JOY_CLOCK);			// Set output to off
	
	// Joystick latch output
	JOY_LATCH_DDR |= JOY_LATCH;				// Set direction to output
	JOY_LATCH_PORT &= ~(JOY_LATCH);			// Set output to off
	
	// Number of players switch
	PLAYERS_DDR &= ~(PLAYERS);				// Set direction to input
	PLAYERS_PORT |= PLAYERS;				// Turn on weak pull-ups
	
	// Turbo select switches
	TURBOA_DDR &= ~(TURBOA);				// Set direction to input
	TURBOA_PORT |= TURBOA;					// Turn on weak pull-ups
	TURBOB_DDR &= ~(TURBOB);				// Set direction to input
	TURBOB_PORT |= TURBOB;					// Turn on weak pull-ups
	
	// Initialise the LUFA USB library
	USB_Init();
}

// Read the physical state of the buttons on the four score
void readButtonStates(void)
{
	// Buttons and switches on the four score (note: only mapped to joystick 0)
	if (PLAYERS_PIN & PLAYERS) joyStick[0].button[MAP_PLAYERS_BUTTON].physicalState = 1;
		else joyStick[0].button[MAP_PLAYERS_BUTTON].physicalState = 0;
	if (TURBOA_PIN & TURBOA) joyStick[0].button[MAP_TURBOA_BUTTON].physicalState = 1;
		else joyStick[0].button[MAP_TURBOA_BUTTON].physicalState = 0;
	if (TURBOB_PIN & TURBOB) joyStick[0].button[MAP_TURBOB_BUTTON].physicalState = 1;
		else joyStick[0].button[MAP_TURBOB_BUTTON].physicalState = 0;
}

// Read the joystick button states for all 4 joysticks
void readJoystickStates(void)
{
	// Set joystick latch low
	JOY_LATCH_PORT &= ~(JOY_LATCH);
	
	for (uint8_t buttonNumber = 0; buttonNumber < NUMBER_OF_BUTTONS - 3; buttonNumber++)
	{
		// Set joystick clock low
		_delay_us(6);
		JOY_CLOCK_PORT &= ~(JOY_CLOCK);
		
		// Read the data pin state for all joysticks
		if (JOY1_DATA_PIN & JOY1_DATA) joyStick[0].button[buttonNumber].physicalState = BUTTON_OFF;
			else joyStick[0].button[buttonNumber].physicalState = BUTTON_ON;
		if (JOY2_DATA_PIN & JOY2_DATA) joyStick[1].button[buttonNumber].physicalState = BUTTON_OFF;
			else joyStick[1].button[buttonNumber].physicalState = BUTTON_ON;
		if (JOY3_DATA_PIN & JOY3_DATA) joyStick[2].button[buttonNumber].physicalState = BUTTON_OFF;
			else joyStick[2].button[buttonNumber].physicalState = BUTTON_ON;
		if (JOY4_DATA_PIN & JOY4_DATA) joyStick[3].button[buttonNumber].physicalState = BUTTON_OFF;
			else joyStick[3].button[buttonNumber].physicalState = BUTTON_ON;
		
		// Set joystick clock high
		_delay_us(6);
		JOY_CLOCK_PORT |= (JOY_CLOCK);
	}
	
	// Set joystick latch high
	JOY_LATCH_PORT |= (JOY_LATCH);
}

// Debounce buttons and joysticks on and off to improve joystick feedback
void performDebounce(void)
{
	for (uint8_t joystickNumber = 0; joystickNumber < 4; joystickNumber++)
	{
		// De-bounce all buttons on and off
		for (uint8_t buttonNumber = 0; buttonNumber < NUMBER_OF_BUTTONS; buttonNumber++)
		{
			// De-bounce on
			if (joyStick[joystickNumber].button[buttonNumber].physicalState ==
				BUTTON_ON && joyStick[joystickNumber].button[buttonNumber].state == BUTTON_OFF)
			{
				// If the de-bounce tolerance is met change state otherwise
				// increment the de-bounce counter
				if (joyStick[joystickNumber].button[buttonNumber].debounceCount > DEBOUNCE_TOLERANCE)
				joyStick[joystickNumber].button[buttonNumber].state = BUTTON_ON;
				else joyStick[joystickNumber].button[buttonNumber].debounceCount++;
			}
			// De-bounce off
			if (joyStick[joystickNumber].button[buttonNumber].physicalState
				== BUTTON_OFF && joyStick[joystickNumber].button[buttonNumber].state == BUTTON_ON)
			{			
				// If the de-bounce tolerance is met change state otherwise
				// increment the de-bounce counter
				if (joyStick[joystickNumber].button[buttonNumber].debounceCount > DEBOUNCE_TOLERANCE)
				joyStick[joystickNumber].button[buttonNumber].state = BUTTON_OFF;
				else joyStick[joystickNumber].button[buttonNumber].debounceCount++;
			}
			// Reset de-bounce counter
			if (joyStick[joystickNumber].button[buttonNumber].physicalState ==
				joyStick[joystickNumber].button[buttonNumber].state)
				joyStick[joystickNumber].button[buttonNumber].debounceCount = 0;
		}
	}		
}

// Event handler for the USB_Connect event
void EVENT_USB_Device_Connect(void)
{
	// Nothing to do here
}

// Event handler for the USB_Disconnect event
void EVENT_USB_Device_Disconnect(void)
{
	// Nothing to do here
}

// Event handler for the USB_ConfigurationChanged event
void EVENT_USB_Device_ConfigurationChanged(void)
{
	// Setup HID Report Endpoints, one for each joystick
	Endpoint_ConfigureEndpoint(JOYSTICK0_EPADDR, EP_TYPE_INTERRUPT, JOYSTICK_EPSIZE, 1);
	Endpoint_ConfigureEndpoint(JOYSTICK1_EPADDR, EP_TYPE_INTERRUPT, JOYSTICK_EPSIZE, 1);
	Endpoint_ConfigureEndpoint(JOYSTICK2_EPADDR, EP_TYPE_INTERRUPT, JOYSTICK_EPSIZE, 1);
	Endpoint_ConfigureEndpoint(JOYSTICK3_EPADDR, EP_TYPE_INTERRUPT, JOYSTICK_EPSIZE, 1);
}

// Event handler for the USB_ControlRequest event
void EVENT_USB_Device_ControlRequest(void)
{
	// Handle HID Class specific requests
	switch (USB_ControlRequest.bRequest)
	{
		case HID_REQ_GetReport:
			if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				USB_JoystickReport_Data_t JoystickReportData;
				
				// Check which joystick the control request refers to:
				switch (USB_ControlRequest.wIndex)
				{
					case 0x00:
						// Create the next joystick0 HID report to send to the host
						GetNextReport(&JoystickReportData, &previousJoystickReportData0, 0);

						Endpoint_ClearSETUP();

						// Write the joystick0 report data to the control endpoint
						Endpoint_Write_Control_Stream_LE(&JoystickReportData, sizeof(JoystickReportData));
						Endpoint_ClearOUT();
						break;

					case 0x01:
						// Create the next joystick1 HID report to send to the host
						GetNextReport(&JoystickReportData, &previousJoystickReportData1, 1);

						Endpoint_ClearSETUP();

						// Write the joystick1 report data to the control endpoint
						Endpoint_Write_Control_Stream_LE(&JoystickReportData, sizeof(JoystickReportData));
						Endpoint_ClearOUT();
						break;
						
					case 0x02:
						// Create the next joystick2 HID report to send to the host
						GetNextReport(&JoystickReportData, &previousJoystickReportData2, 2);

						Endpoint_ClearSETUP();

						// Write the joystick2 report data to the control endpoint
						Endpoint_Write_Control_Stream_LE(&JoystickReportData, sizeof(JoystickReportData));
						Endpoint_ClearOUT();
						break;

					case 0x03:
						// Create the next joystick3 HID report to send to the host
						GetNextReport(&JoystickReportData, &previousJoystickReportData3, 3);

						Endpoint_ClearSETUP();

						// Write the joystick3 report data to the control endpoint
						Endpoint_Write_Control_Stream_LE(&JoystickReportData, sizeof(JoystickReportData));
						Endpoint_ClearOUT();
						break;						
				}				
			}

			break;
	}
}

// Generates the next report to be sent to the host
bool GetNextReport(USB_JoystickReport_Data_t* const reportData, USB_JoystickReport_Data_t* const previousReportData, uint8_t joystickNumber)
{
	bool inputChanged = false;

	// Clear the report contents
	memset(reportData, 0, sizeof(USB_JoystickReport_Data_t));
	
	// Set the joystick button status
	if (joyStick[joystickNumber].button[MAP_FIREA_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 0);
	if (joyStick[joystickNumber].button[MAP_FIREB_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 1);
	if (joyStick[joystickNumber].button[MAP_SELECT_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 2);
	if (joyStick[joystickNumber].button[MAP_START_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 3);
	
	// Check the four score button status
	if (joyStick[joystickNumber].button[MAP_PLAYERS_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 4);
	if (joyStick[joystickNumber].button[MAP_TURBOB_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 5);
	if (joyStick[joystickNumber].button[MAP_TURBOA_BUTTON].state == BUTTON_ON)	reportData->buttonMask |= (1 << 6);
	
	// Set the joystick direction states
	if (joyStick[joystickNumber].button[MAP_UP_BUTTON].state == BUTTON_ON)			reportData->y = 127;
	else if (joyStick[joystickNumber].button[MAP_DOWN_BUTTON].state == BUTTON_ON)	reportData->y = -127;
	else reportData->y = 0;
	
	if (joyStick[joystickNumber].button[MAP_LEFT_BUTTON].state == BUTTON_ON)		reportData->x = -127;
	else if (joyStick[joystickNumber].button[MAP_RIGHT_BUTTON].state == BUTTON_ON)	reportData->x = 127;
	else reportData->x = 0;

	// Check to see if the joystick state has changed since the last report was sent
	if (reportData->buttonMask != previousReportData->buttonMask) inputChanged = true;
	if (reportData->x != previousReportData->x) inputChanged = true;
	if (reportData->y != previousReportData->y) inputChanged = true;
	
	// Save the current joystick status for later comparison
	memcpy(previousReportData, reportData, sizeof(USB_JoystickReport_Data_t));

	// Return whether the new report is different to the previous report or not
	return inputChanged;
}

// Function to manage HID report generation and transmission to the host
void HID_Task(void)
{
	// Device must be connected and configured for the task to run
	if (USB_DeviceState != DEVICE_STATE_Configured)
	  return;

	// Select the Joystick 0 Report Endpoint
	Endpoint_SelectEndpoint(JOYSTICK0_EPADDR);

	// Check to see if the host is ready for another packet
	if (Endpoint_IsINReady())
	{
		USB_JoystickReport_Data_t JoystickReportData0;

		// Create the next HID report to send to the host
		GetNextReport(&JoystickReportData0, &previousJoystickReportData0, 0);

		// Write Joystick Report Data
		Endpoint_Write_Stream_LE(&JoystickReportData0, sizeof(JoystickReportData0), NULL);

		// Finalize the stream transfer to send the last packet
		Endpoint_ClearIN();

		// Clear the report data afterwards
		memset(&JoystickReportData0, 0, sizeof(JoystickReportData0));
	}
	
	// Select the Joystick 1 Report Endpoint
	Endpoint_SelectEndpoint(JOYSTICK1_EPADDR);
	
	// Check to see if the host is ready for another packet
	if (Endpoint_IsINReady())
	{
		USB_JoystickReport_Data_t JoystickReportData1;

		// Create the next HID report to send to the host
		GetNextReport(&JoystickReportData1, &previousJoystickReportData1, 1);

		// Write Joystick Report Data
		Endpoint_Write_Stream_LE(&JoystickReportData1, sizeof(JoystickReportData1), NULL);

		// Finalize the stream transfer to send the last packet
		Endpoint_ClearIN();

		// Clear the report data afterwards
		memset(&JoystickReportData1, 0, sizeof(JoystickReportData1));
	}
	
	// Select the Joystick 2 Report Endpoint
	Endpoint_SelectEndpoint(JOYSTICK2_EPADDR);
		
	// Check to see if the host is ready for another packet
	if (Endpoint_IsINReady())
	{
		USB_JoystickReport_Data_t JoystickReportData2;

		// Create the next HID report to send to the host
		GetNextReport(&JoystickReportData2, &previousJoystickReportData2, 2);

		// Write Joystick Report Data
		Endpoint_Write_Stream_LE(&JoystickReportData2, sizeof(JoystickReportData2), NULL);

		// Finalize the stream transfer to send the last packet
		Endpoint_ClearIN();

		// Clear the report data afterwards
		memset(&JoystickReportData2, 0, sizeof(JoystickReportData2));
	}
		
	// Select the Joystick 1 Report Endpoint
	Endpoint_SelectEndpoint(JOYSTICK3_EPADDR);
			
	// Check to see if the host is ready for another packet
	if (Endpoint_IsINReady())
	{
		USB_JoystickReport_Data_t JoystickReportData3;

		// Create the next HID report to send to the host
		GetNextReport(&JoystickReportData3, &previousJoystickReportData3, 3);

		// Write Joystick Report Data
		Endpoint_Write_Stream_LE(&JoystickReportData3, sizeof(JoystickReportData3), NULL);

		// Finalize the stream transfer to send the last packet
		Endpoint_ClearIN();

		// Clear the report data afterwards
		memset(&JoystickReportData3, 0, sizeof(JoystickReportData3));
	}
}
